// prime.cpp - finds large prime numbers using a probabilistic algorithm

// written by David Levine, based on code by Mark Weiss
//		January 1997


#include <iostream.h>
#include "apstring.h"
#include "dice.h"
#include "bigint.h"

const bool DEBUGGING = false;	// TRUE iff we want to trace the number

BigInt RandomBigInt( int N )
// returns a random N-digit big integer
{
	Dice d(10);  
	
	BigInt Answer(d.Roll()-1);	// Answer has one digit

	for (int i=1; i<N; i++)		// Add N-1 more digits to Answer
	{
		Answer = Answer*10 + (d.Roll()-1);
	}
	return Answer;
}

BigInt
Power(const BigInt & base, const BigInt & expo, const BigInt & N)
// precondition: expo >= 0     
// postcondition: returns (base^expo)%N (base to the power expo mod N)
{
    if (0 == expo)
    {
        return 1;                             // correct for zeroth power
    }
    else
    {
        BigInt semi(Power(base,expo/2,N));    // build answer from this
        BigInt temp((semi*semi) % N);
            
            // k^2 mod a prime is 1 iff k = +/- 1.  
            // If temp and semi violate this, then N is composite!
    
        if ((temp == 1) && (semi != 1) && (semi != N-1)) 
        {
	        if (DEBUGGING)
          {
	    	    cout << "    Killed by " << semi << " squared" << endl;
          }
       	  return 0;			// This is an artificial way of guaranteeing
       	    				// that we later fail the Fermat test (since 0 != 1)
        }
       
    	  if (expo % 2 == 0)                      // even exponent
        {
        	return temp;
        }
    	  else                                    // odd exponent
        {
        	return (base*temp) % N;
        }       
    }
}


bool IsWitness(const BigInt & w, const BigInt & N)
// postcondition: returns TRUE iff w is a witness to N's compositeness
//	This can result either because w and N failing to meet Fermat's
//	Little Theorem (that a^(p-1) is congruent to 1 mod p) or because
//	a non unit value of k was found to have its square congruent to 1 mod N
{
	bool isComposite(false);
	
	if (DEBUGGING)
	{
		cout << "Witness: " << w << "  N: " << N << endl;
	}
	
	BigInt R = Power( w, N-1, N );
	
	if (DEBUGGING)
	{
		if ((R!=1) && (R!=0)) 
      cout << "    Failed Fermat Test" << endl;
	}
	return (isComposite || (R != 1));
}

bool IsPrime(const BigInt & N, int Trials, int size)
// postcondition: Returs TRUE iff Trials witnesses fail to show that N is composite
{
	BigInt witness;
	
	for (int i=0; i<Trials; i++) 
	{
		// Start by finding a random number less than N.  
		
		witness = RandomBigInt(size);	
		while ((witness >= N) || (witness == 0))
		{
			witness = RandomBigInt(size);		
		}	
		// Then check its properties with respect to N
		
		if (IsWitness(witness, N)) 
		{
			return false;
		}
	}
	return true;
}


int main()
{
  char ch;
  int size;
  int numWitness;
	BigInt t; 		// Our (hopeful) prime

  cout << "Enter number of digits: ";
  cin  >> size;
  cout << "Enter number of \"Witnesses\" (20 is a good number): ";
  cin  >> numWitness;
		
	do {
		t = RandomBigInt(size);		// Get a random test prime
		if (t%2 == 0)			        // At least make sure it's odd
		{
			t += 1;
		}
		cout << '.';
	} while (! IsPrime(t,numWitness, size)); 
	
	cout << endl << endl << t << " is (probably) prime." << endl << endl;
	
	return 0;
}

